' File: CopyPasteGroup4.BAS
' Version: Version 4
' Date: January 12, 2015
' Author: Ali Gokce, ATA Engineering
' www.ata-plmsoftware.com
' Description: This API copies FE entities (matl, prop, layup, node & elem) in a group in an open model to
'  another open model without using the harddisk. Numbering conflicts are resolved automatically. 

'Usage Notes
'====================================
' The recipient (destination) model should be active when the API is called. 
' Run rebuild afterwards to make sure that the operation was flawless. 
' Note that coordinate systems are not copied.

'Copyright (c) <2007> <ATA Engineering, Inc.>
'
'Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
'associated documentation files (the "Software") to use, copy, modify, merge, and/or distribute
'royalty-free copies of the Software, and to permit persons to whom the Software is furnished
'to do so, subject to the following conditions:
'The above copyright notice and this permission notice shall be included in all copies or
'substantial portions of the Software.
'THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
'INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
'FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
'WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
'OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Sub Main
    'Define a string indicating the version of the translator
    Const API_version As String = "4"

    Dim App As femap.model
    Set App = GetObject(,"femap.model")

    'Indicate which version and contact information for support
    App.feAppMessage(FCM_NORMAL,"*** Copy Paste Group, Version " & API_version & " ***")
    App.feAppMessage(FCM_NORMAL,"Created by ATA Engineering, http://www.ata-plmsoftware.com")

    Dim NofModels As Long
    Dim ActModelID As Huge_
    Dim ModelIDV As Variant
    Dim modelname As String
    Dim FromModel As Variant

    Dim NodeMap() As Long
	Dim MatMap() As Long
    Dim LayupMap() As Long
    Dim PropMap() As Long

    Dim ElemV() As femap.Elem
    Dim NodeV() As femap.Node
    Dim MatV() As femap.Matl
    Dim LayupV() As femap.Layup
    Dim PropV() As femap.Prop

    Dim elemSET As femap.Set
	Dim nodeSET As femap.Set
	Dim propSET As femap.Set
	Dim layupSET As femap.Set
	Dim matSET As femap.Set
	Dim feGroup As femap.Group

	Dim SortNode As femap.SortSet
    Dim SortMat	As femap.SortSet
    Dim SortLayup As femap.SortSet
    Dim SortProp As femap.SortSet
    Dim SortElem As femap.SortSet

	Set elemSET = App.feSet()
	Set nodeSET = App.feSet()
	Set propSET = App.feSet()
	Set layupSET= App.feSet()
	Set matSET = App.feSet()
	Set feGroup = App.feGroup()

	Set SortNode = App.feSort()
    Set SortMat	= App.feSort()
    Set SortLayup = App.feSort()
    Set SortProp = App.feSort()
	Set SortElem = App.feSort()

    rc = App.feAppGetAllModels(NofModels,ModelIDV)
    If rc = FE_FAIL Then
		App.feAppMessage(3,"No valid open models!... exiting API")
        Exit All
    End If

    rc = App.feAppGetModel(ActModelID)

    App.feAppMessage(0,"C U R R E N T L Y    O P E N     M O D E L S")
    For j=0 To NofModels-1
    	CurrID = ModelIDV(j)
		rc = App.feAppGetModelName(CurrID, modelname)
        If CurrID=ActModelID Then
        	rc = App.feAppMessage(0,"ID : " & CurrID & "    Model Name : " & modelname & "   (**ACTIVE**)")
		Else
        	rc = App.feAppMessage(0,"ID : " & CurrID & "    Model Name : " & modelname)
        End If
    Next

	FromModel = InputBox("Enter the model by ID (see messages window) To import FE Group from", "Import From", "", ,)
	Flag = -1
    If IsNumeric(FromModel) Then
   	    For j=0 To NofModels-1
        	If FromModel = ModelIDV(j) Then
            	Flag = 1
            End If
        Next
        If Flag = -1 Then
			App.feAppMessage(3,"Entered model ID is not valid!... exiting API")
            Exit All
        End If
    Else
		App.feAppMessage(3,"Entered model ID is not a valid ID!... exiting API")
        Exit All
    End If

    App.feAppMessage(0,"---- switching to donor model -------")
    rc = App.feAppSetModel(FromModel)
    If rc = FE_FAIL Then
		App.feAppMessage(3,"Entered model ID is not valid!... exiting API")
        Exit All
    End If

	GroupID = InputBox("Enter the group ID to import to the active model","Group to Import","",300,800)
    If feGroup.Exist(GroupID) = FE_FAIL Then
		App.feAppMessage(3,"Group does not exist!... Exiting API")
        Exit All
    End If

    rc = elemSET.Clear()
    rc = elemSET.AddGroup(FT_ELEM, groupID)
    elemSETID = elemSET.ID
    elemSET.Save = True
    rc = nodeSET.Clear()
    rc = nodeSET.AddGroup(FT_NODE, groupID)
    nodeSETID = nodeSET.ID
    nodeSET.Save = True
    rc = propSET.Clear()
	rc = propSET.AddGroup(FT_PROP, groupID)
    propSETID = propSET.ID
    propSET.Save = True
    rc = matSET.Clear()
    rc = matSET.AddGroup(FT_MATL, groupID)
    matSETID = matSET.ID
    matSET.Save = True
    rc = layupSET.Clear()
	rc = layupSET.AddGroup(FT_LAYUP,groupID)
    layupSETID = layupSET.ID
    layupSET.Save = True
	App.feAppMessage(0, elemSET.count & " elems in group " & groupID & ". IDs from " & elemSET.First() & " to " & elemSET.Last() & ". ")
	App.feAppMessage(0, nodeSET.count & " nodes in group " & groupID & ". IDs from " & nodeSET.First() & " to " & nodeSET.Last() & ". ")
	App.feAppMessage(0, propSET.count & " props in group " & groupID & ". IDs from " & propSET.First() & " to " & propSET.Last() & ". ")
	App.feAppMessage(0, layupSET.count & " layups in group " & groupID & ". IDs from " & layupSET.First() & " to " & layupSET.Last() & ". ")
	App.feAppMessage(0, matSET.count & " matrls in group " & groupID & ". IDs from " & matSET.First() & " to " & matSET.Last() & ". ")

	App.feAppMessage(0, "NodeSETID " & nodeSETID)

    ReDim ElemV(elemSET.count)
	ReDim NodeV(nodeSET.count)
	ReDim PropV(propSET.count)
	ReDim MatV(matSET.count)
	ReDim LayupV(layupSET.count)

    ReDim ElemMap(elemSET.count,2)
	ReDim NodeMap(nodeSET.count,2)
	ReDim PropMap(propSET.count,2)
	ReDim MatMap(matSET.count,2)
	ReDim LayupMap(layupSET.count,2)

	elemSEL = elemSET.First()
    Ecount = 0
	While elemSEL > 0
    	Set ElemV(Ecount)=App.feElem()
    	rc = ElemV(Ecount).Get(elemSEL)
        If rc = -1 Then
			'App.feAppMessage(0, "Node " & NodeV(count).ID  & " read. nodeSel " & nodeSEL & " . Exist? " & rc)
        	Ecount = Ecount+1
            elemSEL = elemSET.Next()
        Else
            DeleteE = elemSEL
            elemSEL = elemSET.Next()
			rc = elemSET.Remove(DeleteE)
		End If
    Wend

	nodeSEL = nodeSET.First()
    Ncount = 0
	While nodeSEL > 0
    	Set NodeV(Ncount)=App.feNode()
    	rc = NodeV(Ncount).Get(nodeSEL)
        If rc = -1 Then
			'App.feAppMessage(0, "Node " & NodeV(count).ID  & " read. nodeSel " & nodeSEL & " . Exist? " & rc)
        	Ncount = Ncount+1
            nodeSEL = nodeSET.Next()
        Else
            DeleteE = nodeSEL
            nodeSEL = nodeSET.Next()
			rc = nodeSET.Remove(DeleteE)
		End If
    Wend

   	propSEL = propSET.First()
    Pcount = 0
	While propSEL > 0
    	Set PropV(Pcount)=App.feProp()
    	rc = PropV(Pcount).Get(propSEL)
        If rc = -1 Then
			'App.feAppMessage(0, "Node " & NodeV(count).ID  & " read. nodeSel " & nodeSEL & " . Exist? " & rc)
        	Pcount = Pcount+1
            propSEL = propSET.Next()
        Else
            DeleteE = propSEL
            propSEL = propSET.Next()
			rc = propSET.Remove(DeleteE)
		End If
    Wend

  	matSEL = matSET.First()
    Mcount = 0
	While matSEL > 0
    	Set MatV(Mcount)=App.feMatl()
    	rc = MatV(Mcount).Get(matSEL)
        If rc = -1 Then
			'App.feAppMessage(0, "Node " & NodeV(count).ID  & " read. nodeSel " & nodeSEL & " . Exist? " & rc)
        	Mcount = Mcount+1
            matSEL = matSET.Next()
        Else
            DeleteE = matSEL
            matSEL = matSET.Next()
			rc = matSET.Remove(DeleteE)
		End If
    Wend


	layupSEL = layupSET.First()
    Lcount = 0
	While layupSEL > 0
    	Set LayupV(Lcount)=App.feLayup()
    	rc = LayupV(Lcount).Get(layupSEL)
        If rc = -1 Then
			'App.feAppMessage(0, "Node " & NodeV(count).ID  & " read. nodeSel " & nodeSEL & " . Exist? " & rc)
        	Lcount = Lcount+1
            layupSEL = layupSET.Next()
        Else
            DeleteE = layupSEL
            layupSEL = layupSET.Next()
			rc = layupSET.Remove(DeleteE)
		End If
    Wend

    rc = App.feAppSetModel(ActModelID)
    If rc = FE_FAIL Then
		App.feAppMessage(3,"Entered model ID is not valid!... exiting API")
        Exit All
    End If

	App.feAppMessage(0,"---- switching to destination (recipient) model -------")

	App.feAppMessage(0, elemSET.count & " elems in group " & groupID & ". IDs from " & elemSET.First() & " to " & elemSET.Last() & ". ")
	App.feAppMessage(0, nodeSET.count & " nodes in group " & groupID & ". IDs from " & nodeSET.First() & " to " & nodeSET.Last() & ". ")
	App.feAppMessage(0, propSET.count & " props in group " & groupID & ". IDs from " & propSET.First() & " to " & propSET.Last() & ". ")
	App.feAppMessage(0, layupSET.count & " layups in group " & groupID & ". IDs from " & layupSET.First() & " to " & layupSET.Last() & ". ")
	App.feAppMessage(0, matSET.count & " matrls in group " & groupID & ". IDs from " & matSET.First() & " to " & matSET.Last() & ". ")

    rc = elemSET.Reload(elemSET.ID)
    rc = nodeSET.Reload(nodeSETID)
    rc = propSET.Reload(propSETID)
    rc = matSET.Reload(matSETID)
    rc = layupSET.Reload(layupSETID)

	App.feAppMessage(0,"---- reloading sets -------")

	App.feAppMessage(0, elemSET.count & " elems in group " & groupID & ". IDs from " & elemSET.First() & " to " & elemSET.Last() & ". ")
	App.feAppMessage(0, nodeSET.count & " nodes in group " & groupID & ". IDs from " & nodeSET.First() & " to " & nodeSET.Last() & ". ")
	App.feAppMessage(0, propSET.count & " props in group " & groupID & ". IDs from " & propSET.First() & " to " & propSET.Last() & ". ")
	App.feAppMessage(0, layupSET.count & " layups in group " & groupID & ". IDs from " & layupSET.First() & " to " & layupSET.Last() & ". ")
	App.feAppMessage(0, matSET.count & " matrls in group " & groupID & ". IDs from " & matSET.First() & " to " & matSET.Last() & ". ")

	App.feAppMessage(0, "NodeSETID " & NodeSETID)

    For j=0 To Ncount-1
    	OldID = NodeV(j).ID
    	rc = NodeV(j).Exist(OldID)
    	If  rc = FE_FAIL Then
        	NewID = NodeV(j).ID
       	    rc = NodeV(j).Put(0)
            'App.deAppMessage(0, " node : " & OldID & " written as " & NewID)
        Else
        	NewID = NodeV(j).NextEmptyID
        	rc = NodeV(j).Put(NewID)
            'App.deAppMessage(0, " node : " & OldID & " written as " & NewID)
        End If
        SortNode.nInt1 = NewID
        rc = SortNode.Add(OldID)
       'App.feAppMessage(2," OldID= " & OldID & " NewID = " & NewID &  " SortNode.nInt1=" & SortNode.nInt1)
    Next
    App.feAppMessage(0, " --------- nodes imported -------------")

    For j=0 To Mcount-1
    	OldID = MatV(j).ID
    	rc = MatV(j).Exist(OldID)
    	If  rc = FE_FAIL Then
        	NewID = MatV(j).ID
       	    rc = MatV(j).Put(0)
            'App.deAppMessage(0, " Mat : " & OldID & " written as " & NewID)
        Else
        	NewID = MatV(j).NextEmptyID
        	rc = MatV(j).Put(NewID)
            'App.deAppMessage(0, " Mat : " & OldID & " written as " & NewID)
        End If
        SortMat.nInt1 = NewID
        rc = SortMat.Add(OldID)
    Next
    App.feAppMessage(0, " --------- materials imported -------------")

    For j=0 To Lcount-1
    	OldID = LayupV(j).ID
        For k=0 To LayupV(j).NumberOfPlys-1
			rc = SortMat.Find(LayupV(j).matlID(k))
            If rc = FE_OK Then
   			   LayupV(j).matlID(k) = SortMat.nInt1
            End If
		Next
    	rc = LayupV(j).Exist(LayupV(j).ID)
    	If  rc = FE_FAIL Then
        	NewID = LayupV(j).ID
       	    rc = LayupV(j).Put(0)
            'App.deAppMessage(0, " Layup : " & OldID & " written as " & NewID)
        Else
        	NewID = LayupV(j).NextEmptyID
        	rc = LayupV(j).Put(NewID)
            'App.deAppMessage(0, " Layup : " & OldID & " written as " & NewID)
        End If
        SortLayup.nInt1 = NewID
        rc = SortLayup.Add(OldID)
    Next
    App.feAppMessage(0, " --------- layups imported -------------")

    For j=0 To Pcount-1
    	OldID=PropV(j).ID
        If PropV(j).type=FET_L_LAMINATE_PLATE Or PropV(j).type=FET_P_LAMINATE_PLATE Then
        	rc = SortLayup.Find(PropV(j).layupID)
            If rc = FE_OK Then
               PropV(j).layupID = SortLayup.nInt1
            End If
         Else
         	rc = SortMat.Find(PropV(j).matlID)
            If rc = FE_OK Then
               PropV(j).matlID = SortMat.nInt1
            End If
            For k=0 To 9
            	rc = SortMat.Find(PropV(j).ExtraMatlID(k))
                If rc = FE_OK Then
                   PropV(j).ExtraMatlID(k) = SortMat.nInt1
                End If
           Next
        End If
    	rc = PropV(j).Exist(OldID)
    	If  rc = FE_FAIL Then
        	NewID = PropV(j).ID
       	    rc = PropV(j).Put(0)
            'App.deAppMessage(0, " Prop : " & OldID & " written as " & NewID)
       Else
        	NewID = PropV(j).NextEmptyID
        	rc = PropV(j).Put(NewID)
            'App.deAppMessage(0, " Prop : " & OldID & " written as " & NewID)
        End If
        SortProp.nInt1 = NewID
        rc = SortProp.Add(OldID)
    Next
    App.feAppMessage(0, " --------- properties imported -------------")

	App.feAppMessage(0, " --------- importing elements - see the STATUS bar below -------------")
    App.feAppStatusShow(True, Ecount)
    Dim nodecount As Long
    For j=0 To Ecount-1
    	OldID=ElemV(j).ID
        If ElemV(j).type = FET_L_RIGID Then
           	OldNode = ElemV(j).Nodes(0)
            rc = SortNode.Find(OldNode)
            If rc = FE_OK Then
               ElemV(j).Node(0) = SortNode.nInt1
            End If
            rc = ElemV(j).GetNodeList(0, nodecount, NodeL, face, weight, dof )
			For k=0 To nodecount-1
            	rc = SortNode.Find(NodeL(k))
                If rc = FE_OK Then
                   NodeL(k)=SortNode.nInt1
                End If
            Next
            rc = ElemV(j).PutNodeList(0, nodecount, NodeL, face, weight, dof )
        Else
			rc = SortProp.Find(ElemV(j).propID)
            'App.feAppMessage(FCM_NORMAL, "rc =  " & rc & ", ElemV(j).propID =" & ElemV(j).propID)
            If rc = FE_OK Then
              ' rc= App.feAppMessage(FCM_NORMAL, "NABER ")
               ElemV(j).propID = SortProp.nInt1
            End If
            For k=0 To UBound(ElemV(j).Nodes)
            	OldNID = ElemV(j).Node(k)
                If OldNID > 0 Then
            		rc = SortNode.Find(OldNID)
                	If rc = FE_OK Then
              			ElemV(j).Node(k) = SortNode.nInt1
                	End If
                	'App.feAppMessage(2," Elem: k=" & k & " Old Node ID " & OldNID & " rc = " & rc &  " SortNode.nInt1=" & SortNode.nInt1 & " ElemV(j).Nodes(k)= " & ElemV(j).Nodes(k)  )
                End If
            Next
		End If
       	rc = ElemV(j).Exist(OldID)
    	If  rc = FE_FAIL Then
        	NewID = ElemV(j).ID
       	    rc = ElemV(j).Put(0)
            'App.feAppMessage(0, " Elem : " & OldID & " written as " & NewID)
        Else
        	NewID = ElemV(j).NextEmptyID
        	rc = ElemV(j).Put(NewID)
            'App.feAppMessage(0, " Elem : " & OldID & " written as " & NewID)
        End If
        SortElem.nInt1 = NewID 'it may be needed
        rc = SortElem.Add(OldID)
        App.feAppStatusUpdate(j)
    Next
    App.feAppStatusShow(False, Ecount)
    App.feAppMessage(0, " --------- elements imported -------------")

    groupID = feGroup.NextEmptyID
    feGroup.ID = groupID

    rc = SortNode.Reset()
    rc = nodeSET.Clear()
    While SortNode.Next()
    	rc = nodeSET.Add(SortNode.nInt1)
    Wend

    rc = SortElem.Reset()
	rc = elemSET.Clear()
    While SortElem.Next()
    	rc = elemSET.Add(SortElem.nInt1)
    Wend

    rc = feGroup.SetAdd(FT_NODE,nodeSET.ID)
    rc = feGroup.SetAdd(FT_ELEM,elemSET.ID)
    feGroup.title = "API_Imported_nodes+elems"
    rc = feGroup.Put(groupID)

    App.feAppMessage(2, " Imported items placed in group " & groupID)

End Sub
